import pkg_resources

from . import bloomfilter
from . import identifier_types
from . import randomnames

try:
    __version__ = pkg_resources.get_distribution('clkhash').version
except pkg_resources.DistributionNotFound:
    __version__ = "development"

__author__ = 'N1 Analytics'



from typing import Tuple, Union, Optional, Sequence, Any

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.backends import default_backend


"""
We use the block-size of SHA1 and MD5 as the default key size for HMAC
"""
DEFAULT_KEY_SIZE = 64


class HKDFconfig:
    supported_hash_algos = 'SHA256', 'SHA512'

    def __init__(self, master_secret,   # type: bytes
                 salt=None,             # type: Optional[bytes]
                 info=None,             # type: Optional[bytes]
                 hash_algo='SHA256'     # type: str
                 ):                     # type: (...) -> None
        """
        The parameters for the HDKF are defined as follows:

        master_secret
            input keying material

        salt
            HKDF is defined to operate with and without random salt.  This is
            done to accommodate applications where a salt value is not available.
            We stress, however, that the use of salt adds significantly to the
            strength of HKDF, ensuring independence between different uses of the
            hash function, supporting "source-independent" extraction, and
            strengthening the analytical results that back the HKDF design.

            Random salt differs fundamentally from the initial keying material in
            two ways: it is non-secret and can be re-used.

            Ideally, the salt value is a random (or pseudorandom) string of the
            length HashLen.  Yet, even a salt value of less quality (shorter in
            size or with limited entropy) may still make a significant
            contribution to the security of the output keying material.

        info
            While the 'info' value is optional in the definition of HKDF, it is
            often of great importance in applications.  Its main objective is to
            bind the derived key material to application- and context-specific
            information.  For example, 'info' may contain a protocol number,
            algorithm identifiers, user identities, etc.  In particular, it may
            prevent the derivation of the same keying material for different
            contexts (when the same input key material (IKM) is used in such
            different contexts).  It may also accommodate additional inputs to
            the key expansion part, if so desired (e.g., an application may want
            to bind the key material to its length L, thus making L part of the
            'info' field).  There is one technical requirement from 'info': it
            should be independent of the input key material value IKM.

        hash_algo
            The hash function used by HKDF for the internal HMAC calls. The
            choice of hash function defines the maximum length of the output
            key material. Output bytes <= 255 * hash digest size (in bytes).


        :param master_secret: the input keying material for the kdf as bytes
        :param salt: optional salt value (a non-secret random value) as bytes
        :param info: optional context and application specific information (can be a zero-length string) as bytes
        :param hash_algo: the type of hash function to be used. Either 'SHA256' or 'SHA512'.
        """
        self.master_secret = self.check_is_bytes(master_secret)
        self.info = self.check_is_bytes_or_none(info)
        self.salt = self.check_is_bytes_or_none(salt)
        if hash_algo in HKDFconfig.supported_hash_algos:
            self.hash_algo = hash_algo
        else:
            raise ValueError('hash algorithm "{}" is not supported. Has to be one of {}'.format(hash_algo,
                                                                                                HKDFconfig.supported_hash_algos))

    @staticmethod
    def check_is_bytes(value):
        # type: (Any) -> bytes
        if isinstance(value, bytes):
            return value
        else:
            raise TypeError('provided value is not of type "bytes"')

    @staticmethod
    def check_is_bytes_or_none(value):
        # type: (Any) -> Optional[bytes]
        if value is None:
            return value
        else:
            return HKDFconfig.check_is_bytes(value)


def hkdf(hkdf_config, num_keys, key_size=DEFAULT_KEY_SIZE):
    # type: (HKDFconfig, int, int) -> Tuple[bytes, ...]
    """
    Executes the HKDF key derivation function as described in rfc5869 to derive
    `num_keys` keys of size `key_size` from the master_secret.

    :param hkdf_config: an HKDFconfig object containing the configuration for the HKDF.
    :param num_keys: the number of keys the kdf should produce
    :param key_size: the size of the produced keys
    :return: Derived keys
    """
    hash_dict = {
        'SHA256': hashes.SHA256,
        'SHA512': hashes.SHA512
    }
    if not isinstance(hkdf_config, HKDFconfig):
        raise TypeError('provided config has to be of type "HKDFconfig"')
    hkdf = HKDF(algorithm=hash_dict[hkdf_config.hash_algo](), length=num_keys * key_size, salt=hkdf_config.salt,
                info=hkdf_config.info, backend=default_backend())
    # hkdf.derive returns a block of num_keys * key_size bytes which we divide up into num_keys chunks,
    # each of size key_size
    keybytes = hkdf.derive(hkdf_config.master_secret)
    keys = tuple([keybytes[i * key_size:(i + 1) * key_size] for i in range(num_keys)])
    return keys


def generate_key_lists(master_secrets,              # type: Sequence[Union[bytes, str]]
                       num_identifier,              # type: int
                       key_size=DEFAULT_KEY_SIZE,   # type: int
                       salt=None,                   # type: Optional[bytes]
                       info=None,                   # type: Optional[bytes]
                       kdf='HKDF'                   # type: str
                       ):
    # type: (...) -> Tuple[Tuple[bytes, ...], ...]
    """
    Generates a derived key for each identifier for each master secret using a key derivation function (KDF).

    The only supported key derivation function for now is 'HKDF'.

    The previous key usage can be reproduced by setting kdf to 'legacy'.
    This is highly discouraged, as this strategy will map the same n-grams in different identifier
    to the same bits in the Bloom filter and thus does not lead to good results.

    :param master_secrets: a list of master secrets (either as bytes or strings)
    :param num_identifier: the number of identifiers
    :param key_size: the size of the derived keys
    :param salt: salt for the KDF as bytes
    :param info: optional context and application specific information as bytes
    :param kdf: the key derivation function algorithm to use
    :return: The derived keys.
             First dimension is of size num_identifier, second dimension is the same as master_secrets.
             A key is represented as bytes.
    """
    keys = []
    try:
        for key in master_secrets:
            if isinstance(key, bytes):
                keys.append(key)
            else:
                keys.append(key.encode('UTF-8'))
    except AttributeError:
        raise TypeError("provided 'master_secrets' have to be either of type bytes or strings.")
    if kdf is 'HKDF':
        key_lists = [hkdf(HKDFconfig(key, salt, info), num_identifier, key_size) for key in keys]
        # regroup such that we get a tuple of keys for each identifier
        return tuple(zip(*key_lists))
    if kdf is 'legacy':
        return tuple([tuple(keys) for _ in range(num_identifier)])
    raise ValueError('kdf: "{}" is not supported.'.format(kdf))



from typing import List, Dict, Optional, TextIO

import os
from clkhash.identifier_types import IdentifierType, identifier_type_from_description


def load_schema(schema_file):
    # type: (Optional[TextIO]) -> List[Dict[str, str]]
    if schema_file is None:
        schema = [
            {"identifier": 'INDEX'},
            {"identifier": 'NAME freetext'},
            {"identifier": 'DOB YYYY/MM/DD'},
            {"identifier": 'GENDER M or F'}
        ]
    else:
        filename, extension = os.path.splitext(schema_file.name)

        if extension == '.json':
            import json
            schema = json.load(schema_file)
        elif extension == '.yaml':
            import yaml
            schema = yaml.load(schema_file)
        else:
            schema_line = schema_file.read().strip()
            schema = [{"identifier": s.strip()} for s in schema_line.split(",")]

    return schema


def get_schema_types(schema):
    # type: (List[Dict[str, str]]) -> List[IdentifierType]
    schema_types = [identifier_type_from_description(column) for column in schema]
    return schema_types




from __future__ import print_function
from timeit import default_timer as timer

import os
import tempfile
from clkhash.randomnames import NameList
from clkhash.schema import load_schema, get_schema_types
from clkhash.clk import generate_clk_from_csv


def compute_hash_speed(n, quiet=False):
    # type: (int, bool) -> float
    """
    Hash time.
    """
    namelist = NameList(n)

    os_fd, tmpfile_name = tempfile.mkstemp(text=True)

    with open(tmpfile_name, 'wt') as f:
        f.write("header row\n")
        for person in namelist.names:
            print(','.join([str(field) for field in person]), file=f)

    schema = get_schema_types(load_schema(None))

    with open(tmpfile_name, 'rt') as f:
        start = timer()
        generate_clk_from_csv(f, ('key1', 'key2'), schema, progress_bar=not quiet)
        end = timer()

    os.close(os_fd)
    os.remove(tmpfile_name)

    elapsed_time = end - start
    print("{:6d} hashes in {:.6f} seconds. {:.2f} KH/s".format(n, elapsed_time, n/(1000*elapsed_time)))
    return n / elapsed_time


if __name__ == '__main__':
    for n in [100, 1000, 10000, 50000, 100000]:
        compute_hash_speed(n, quiet=n<=10000)



"""
Generate CLK from CSV file
"""
from tqdm import tqdm

import csv
import logging
import time
import platform
import sys

from typing import List, Any, Iterable, TypeVar, TextIO, Tuple, Union, Sequence, \
    Callable, Optional

import concurrent.futures

from clkhash.bloomfilter import stream_bloom_filters, calculate_bloom_filters, serialize_bitarray
from clkhash.key_derivation import generate_key_lists
from clkhash.identifier_types import IdentifierType
from clkhash.stats import OnlineMeanVariance

log = logging.getLogger('clkhash.clk')


def hash_and_serialize_chunk(chunk_pii_data, # type: Iterable[Tuple[Any]]
                             schema_types,   # type: Iterable[IdentifierType]
                             keys,           # type: Tuple[Tuple[bytes, ...]]
                             xor_folds       # type: int
                             ):
    # type: (...) -> Tuple[List[str], List[int]]
    """
    Generate Bloom filters (ie hash) from chunks of PII then serialize
    the generated Bloom filters. It also computes and outputs the Hamming weight (or popcount) -- the number of bits
    set to one -- of the generated Bloom filters.

    :param chunk_pii_data: An iterable of indexable records.
    :param schema_types: An iterable of identifier type names.
    :param keys: A tuple of two lists of secret keys used in the HMAC.
    :param xor_folds: Number of XOR folds to perform. Each fold halves
        the hash length.
    :return: A list of serialized Bloom filters and a list of corresponding popcounts
    """
    clk_data = []
    clk_popcounts = []
    for clk in stream_bloom_filters(chunk_pii_data, schema_types,
                                    keys, xor_folds):
        clk_data.append(serialize_bitarray(clk[0]).strip())
        clk_popcounts.append(clk[2])
    return clk_data, clk_popcounts


def generate_clk_from_csv(input,             # type: TextIO
                          keys,              # type: Tuple[Union[bytes, str], Union[bytes, str]]
                          schema_types,      # type: List[IdentifierType]
                          no_header=False,   # type: bool
                          progress_bar=True, # type: bool
                          xor_folds=0        # type: int
                          ):
    # type: (...) -> List[str]
    log.info("Hashing data")

    # Read from CSV file
    reader = csv.reader(input)

    # Get the headers
    if not no_header:
        header = input.readline()
        log.info("Header Row: {}".format(header))

    start_time = time.time()

    # Read the lines in CSV file and add it to PII
    pii_data = []
    for line in reader:
        if len(line) == len(schema_types):
            pii_data.append(tuple([element.strip() for element in line]))
        else:
            raise ValueError("Line had unexpected number of elements. " 
                "Expected {} but there was {}".format(
                len(schema_types), len(line)))

    # generate two keys for each identifier
    key_lists = generate_key_lists(keys, len(schema_types))
    if progress_bar:
        stats = OnlineMeanVariance()
        with tqdm(desc="generating CLKs", total=len(pii_data), unit='clk', unit_scale=True,
                  postfix={'mean': stats.mean(), 'std': stats.std()}) as pbar:
            def progress_bar_callback(tics, clk_stats):
                stats.update(clk_stats)
                pbar.set_postfix(mean=stats.mean(), std=stats.std(), refresh=False)
                pbar.update(tics)

            results = generate_clks(pii_data, schema_types, key_lists,
                                    xor_folds, progress_bar_callback)
    else:
        results = generate_clks(pii_data, schema_types, key_lists, xor_folds)

    log.info("Hashing took {:.2f} seconds".format(time.time() - start_time))
    return results


def generate_clks(pii_data,         # type: Sequence[Tuple[str, ...]]
                  schema_types,     # type: List[IdentifierType]
                  key_lists,        # type: Tuple[Tuple[bytes, ...], ...]
                  xor_folds,        # type: int
                  callback=None     # type: Optional[Callable[[int, List[int]], None]]
                  ):
    # type: (...) -> List[Any]
    results = []

    # Chunks PII
    log.info("Hashing {} entities".format(len(pii_data)))
    chunk_size = 200 if len(pii_data) <= 10000 else 1000
    futures = []

    stats = OnlineMeanVariance()

    # Compute Bloom filter from the chunks and then serialise it
    with concurrent.futures.ProcessPoolExecutor() as executor:
        for chunk in chunks(pii_data, chunk_size):
            future = executor.submit(
                hash_and_serialize_chunk,
                chunk, schema_types, key_lists, xor_folds)
            if callback is not None:
                future.add_done_callback(lambda f: callback(len(f.result()[0]), f.result()[1]))
            futures.append(future)

        for future in futures:
            clks, clk_stats = future.result()
            stats.update(clk_stats)
            results.extend(clks)

    return results


T = TypeVar('T')      # Declare generic type variable


def chunks(l, n):
    # type: (Sequence[T], int) -> Iterable[Sequence[T]]
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]



"""
Module to produce a dataset of names, genders and dates of birth and manipulate that list

Currently very simple and not realistic. Additional functions for manipulating the list of names
- producing reordered and subset lists with a specific overlap

ClassList class - generate a list of length n of [id, name, dob, gender] lists

TODO: Get age distribution right by using a mortality table
TODO: Get first name distributions right by using distributions
TODO: Generate realistic errors
TODO: Add RESTfull api to generate reasonable name data as requested
"""
from __future__ import print_function

import csv
from typing import List, Tuple, Iterable, TextIO, Union, Dict

import random
from datetime import datetime, timedelta
import math

import pkgutil

from clkhash.schema import get_schema_types


def load_csv_data(resource_name):
    # type: (str) -> List[str]
    """Loads a specified CSV data file and returns the first column as a Python list
    """
    data_bytes = pkgutil.get_data('clkhash', 'data/{}'.format(resource_name))
    if data_bytes is None:
        raise ValueError("No data resource found with name {}".format(resource_name))
    else:
        data = data_bytes.decode('utf8')
        reader = csv.reader(data.splitlines())
        next(reader, None)  # skip the headers
        return [row[0] for row in reader]


def save_csv(data,      # type: Iterable[Tuple[Union[str, int], ...]]
             schema,    # type: Iterable[Dict[str, str]]
             file       # type: TextIO
             ):
    # type: (...) -> None
    """
    Output generated data to file as CSV with header.

    :param data: An iterable of tuples containing raw data.
    :param schema: Iterable of schema definition dicts
    :param file: A writeable stream in which to write the CSV
    """

    print(','.join(col['identifier'] for col in schema), file=file)
    writer = csv.writer(file)
    writer.writerows(data)


def random_date(start, end):
    # type: (datetime, datetime) -> datetime
    """
    This function will return a random datetime between two datetime objects.

    :param start: datetime of start
    :param end: datetime of end
    :return: random datetime between start and end
    """
    delta = end - start
    int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
    random_second = random.randrange(int_delta)
    return start + timedelta(seconds=random_second)


class NameList:
    """List of randomly generated names"""

    schema = [
            {"identifier": "INDEX"},
            {"identifier": "NAME freetext"},
            {"identifier": "DOB YYYY/MM/DD"},
            {"identifier": "GENDER M or F"}
        ]

    def __init__(self, n):
        # type: (int) -> None
        self.load_names()

        self.earliest_birthday = datetime(year=1916, month=1, day=1)
        self.latest_birthday = datetime(year=2016, month=1, day=1)

        self.names = [person for person in self.generate_random_person(n)]

    @property
    def schema_types(self):
        return get_schema_types(self.schema)

    def generate_random_person(self, n):
        # type: (int) -> Iterable[Tuple[int, str, str, str]]
        """
        Generator that yields details on a person with plausible name, sex and age.

        :yields: Generated data for one person
            tuple - (id: int, name: str('First Last'), birthdate: str('DD/MM/YYYY'), sex: str('M' | 'F') )
        """
        for i in range(n):
            sex = 'M' if random.random() > 0.5 else 'F'
            dob = random_date(self.earliest_birthday, self.latest_birthday).strftime("%Y/%m/%d")
            first_name = random.choice(self.all_male_first_names) if sex == 'M' else random.choice(self.all_female_first_names)
            last_name = random.choice(self.all_last_names)

            yield (
                i,
                first_name + ' ' + last_name,
                dob,
                sex
            )

    def load_names(self):
        # type: () -> None
        """
        This function loads a name database into globals firstNames and lastNames

        initial version uses data files from
        http://www.quietaffiliate.com/free-first-name-and-last-name-databases-csv-and-sql/

        """

        self.all_male_first_names = load_csv_data('male-first-names.csv')
        self.all_female_first_names = load_csv_data('female-first-names.csv')
        self.all_last_names = load_csv_data('CSV_Database_of_Last_Names.csv')

    def generate_subsets(self, sz, overlap=0.8):
        """
        Return a pair of random subsets of the name list with a specified
        proportion of elements in common.

        :param sz: length of subsets to generate
        :param overlap: fraction of the subsets that should have the same names in them
        :raises ValueError: if there aren't sufficiently many names in the list to satisfy
            the request; more precisely, raises if sz + (1 - overlap)*sz > n = len(self.names)
        :return: pair of subsets
        """
        notoverlap = sz - int(math.floor(overlap * sz))
        total_sz = sz + notoverlap
        if total_sz > len(self.names):
            raise ValueError('Requested subset size and overlap demands more '
                             + 'than the number of available names')
        sset = random.sample(self.names, total_sz)
        return sset[:sz], sset[notoverlap:]



#!/usr/bin/env python3

"""
Generate a Bloom filter
"""
from typing import Tuple, Any, Iterable, List

import base64
import hmac
import math
import struct

from future.builtins import range
from enum import Enum
from clkhash.identifier_types import IdentifierType
from functools import partial
from hashlib import sha1, md5

try:
    from hashlib import blake2b
except ImportError:
    from pyblake2 import blake2b  # type: ignore

from bitarray import bitarray


def double_hash_encode_ngrams(ngrams,          # type: Iterable[str]
                              keys,            # type: Tuple[bytes, ...]
                              k,               # type: int
                              l                # type: int
                              ):
    # type: (...) -> bitarray
    """
    Computes the double hash encoding of the provided ngrams with the given keys.

    Using the method from
    http://www.record-linkage.de/-download=wp-grlc-2011-02.pdf

    :param ngrams: list of n-grams to be encoded
    :param keys: hmac secret keys for md5 and sha1 as bytes
    :param k: number of hash functions to use per element of the ngrams
    :param l: length of the output bitarray

    :return: bitarray of length l with the bits set which correspond to the encoding of the ngrams
    """
    key_sha1, key_md5 = keys
    bf = bitarray(l)
    bf.setall(False)
    for m in ngrams:
        sha1hm = int(hmac.new(key_sha1, m.encode(), sha1).hexdigest(), 16) % l
        md5hm = int(hmac.new(key_md5, m.encode(), md5).hexdigest(), 16) % l
        for i in range(k):
            gi = (sha1hm + i * md5hm) % l
            bf[gi] = 1
    return bf


def double_hash_encode_ngrams_non_singular(ngrams,          # type: Iterable[str]
                              keys,            # type: Tuple[bytes, ...]
                              k,               # type: int
                              l                # type: int
                              ):
    # type: (...) -> bitarray.bitarray
    """
    computes the double hash encoding of the provided n-grams with the given keys.

    The original construction of [Schnell2011]_ displays an abnormality for certain inputs:
      An n-gram can be encoded into just one bit irrespective of the number of k.

    Their construction goes as follows: the :math:`k` different indices :math:`g_i` of the Bloom filter for an n-gram
    :math:`x` are defined as:

    .. math::
      g_{i}(x) = (h_1(x) + i h_2(x)) \mod l

    with :math:`0 \leq i < k` and :math:`l` is the length of the Bloom filter. If the value of the hash of :math:`x` of
    the second hash function is a multiple of :math:`l`, then

    .. math::
      h_2(x) = 0 \mod l

    and thus

    .. math::
      g_i(x) = h_1(x) \mod l,

    irrespective of the value :math:`i`. A discussion of this potential flaw can be found
    `here <https://github.com/n1analytics/clkhash/issues/33>`_.

    :param ngrams: list of n-grams to be encoded
    :param key_sha1: hmac secret keys for sha1 as bytes
    :param key_md5: hmac secret keys for md5 as bytes
    :param k: number of hash functions to use per element of the ngrams
    :param l: length of the output bitarray

    :return: bitarray of length l with the bits set which correspond to the encoding of the ngrams
    """
    key_sha1, key_md5 = keys
    bf = bitarray(l)
    bf.setall(False)
    for m in ngrams:
        sha1hm = int(hmac.new(key_sha1, m.encode(), sha1).hexdigest(), 16) % l
        md5hm = int(hmac.new(key_md5, m.encode(), md5).hexdigest(), 16) % l
        i = 0
        while md5hm == 0:
            md5hm = int(hmac.new(key_md5, m.encode() + chr(i).encode(), md5).hexdigest(), 16) % l
            i += 1
        for i in range(k):
            gi = (sha1hm + i * md5hm) % l
            bf[gi] = 1
    return bf


def blake_encode_ngrams(ngrams,          # type: Iterable[str]
                       key,              # type: bytes
                       k,                # type: int
                       l                 # type: int
                       ):
    # type: (...) -> bitarray.bitarray
    """
    Computes the encoding of the provided ngrams using the BLAKE2 hash function.

    We deliberately do not use the double hashing scheme as proposed in [Schnell2011]_, because this
    would introduce an exploitable structure into the Bloom filter. For more details on the
    weakness, see [Kroll2015]_.

    In short, the double hashing scheme only allows for :math:`l^2` different encodings for any possible n-gram,
    whereas the use of :math:`k` different independent hash functions gives you :math:`\sum_{j=1}^{k}{\\binom{l}{j}}`
    combinations.


    **Our construction**

    It is advantageous to construct Bloom filters using a family of hash functions with the property of
    `k-independence <https://en.wikipedia.org/wiki/K-independent_hashing>`_ to compute the indices for an entry.
    This approach minimises the change of collisions.

    An informal definition of *k-independence* of a family of hash functions is, that if selecting a function at random
    from the family, it guarantees that the hash codes of any designated k keys are independent random variables.

    Our construction utilises the fact that the output bits of a cryptographic hash function are uniformly distributed,
    independent, binary random variables (well, at least as close to as possible. See [Kaminsky2011]_ for an analysis).
    Thus, slicing the output of a cryptographic hash function into k different slices gives you k independent random
    variables.

    We chose Blake2 as the cryptographic hash function mainly for two reasons:

    * it is fast.
    * in keyed hashing mode, Blake2 provides MACs with just one hash function call instead of the
      two calls in the HMAC construction used in the double hashing scheme.


    .. warning::
       Please be aware that, although this construction makes the attack of [Kroll2015]_ infeasible, it is most likely
       not enough to ensure security. Or in their own words:

         | However, we think that using independent hash functions alone will not be sufficient to ensure security, since
           in this case other approaches (maybe related to or at least inspired through work from the area of Frequent
           Itemset Mining) are promising to detect at least the most frequent atoms automatically.

    :param ngrams: list of n-grams to be encoded
    :param key: secret key for blake2 as bytes
    :param k: number of hash functions to use per element of the ngrams
    :param l: length of the output bitarray (has to be a power of 2)

    :return: bitarray of length l with the bits set which correspond to the encoding of the ngrams
    """
    log_l = int(math.log(l, 2))
    if not 2**log_l == l:
        raise ValueError('parameter "l" has to be a power of two for the BLAKE2 encoding, but was: {}'.format(l))
    bf = bitarray(l)
    bf.setall(False)
    if k < 1:
        return bf
    num_macs = (k+31) // 32

    for m in ngrams:
        random_shorts = []  # type: List[int]
        for i in range(num_macs):
            hash_bytes = blake2b(m.encode(), key=key, salt=str(i).encode()).digest()
            random_shorts.extend(struct.unpack('32H', hash_bytes))  # interpret hash bytes as 32 unsigned shorts.
        for i in range(k):
            idx = random_shorts[i] % l
            bf[idx] = 1
    return bf


class NgramEncodings(Enum):
    """
    Lists the available schemes for encoding n-grams.

    .. note::
      the slightly awkward looking construction with the calls to partial and the overwrite of __call__ are due to
      compatibility issues with Python 2.7.
    """
    DOUBLE_HASH = partial(double_hash_encode_ngrams)
    """ the initial encoding scheme as described in Schnell, R., Bachteler, T., & Reiher, J. (2011). A Novel 
    Error-Tolerant Anonymous Linking Code. Also see :meth:`double_hash_encode_ngrams`"""
    BLAKE_HASH = partial(blake_encode_ngrams)
    """ uses the BLAKE2 hash function, which is one of the fastest modern hash functions, and does less hash function 
    calls compared to the DOUBLE_HASH based schemes. It avoids one of the exploitable weaknesses of the DOUBLE_HASH 
    scheme. Also see :meth:`blake_encode_ngrams`"""
    DOUBLE_HASH_NON_SINGULAR = partial(double_hash_encode_ngrams_non_singular)
    """ very similar to DOUBLE_HASH, but avoids singularities in the encoding. Also see 
    :meth:`double_hash_encode_ngrams_non_singular`"""

    def __call__(self, *args):
        return self.value(*args)


def fold_xor(bloomfilter,  # type: bitarray
             folds         # type: int
             ):
    # type: (...) -> bitarray
    """ Performs XOR folding on a Bloom filter.

        If the length of the original Bloom filter is n and we perform
        r folds, then the length of the resulting filter is n / 2 ** r.

        :param bloomfilter: Bloom filter to fold
        :param folds: number of folds

        :return: folded bloom filter
    """

    if len(bloomfilter) % 2 ** folds != 0:
        msg = ('The length of the bloom filter is {length}. It is not '
               'divisible by 2 ** {folds}, so it cannot be folded {folds} '
               'times.'
               .format(length=len(bloomfilter), folds=folds))
        raise ValueError(msg)

    for _ in range(folds):
        bf1 = bloomfilter[:len(bloomfilter) // 2]
        bf2 = bloomfilter[len(bloomfilter) // 2:]

        bloomfilter = bf1 ^ bf2

    return bloomfilter


def crypto_bloom_filter(record,       # type: Tuple[Any, ...]
                        tokenizers,   # type: Iterable[IdentifierType]
                        keys,         # type: Tuple[Tuple[bytes, ...]]
                        xor_folds=0,  # type: int
                        l=1024,       # type: int
                        k=30,         # type: int
                        ngram_encoding=NgramEncodings.DOUBLE_HASH
                        ):
    # type: (...) -> Tuple[bitarray, int, int]
    """
    Makes a Bloom filter from a record with given tokenizers and lists of keys.

    Using the method from
    http://www.record-linkage.de/-download=wp-grlc-2011-02.pdf

    :param record: plaintext record tuple. E.g. (index, name, dob, gender)
    :param tokenizers: A list of IdentifierType tokenizers (one for each record element)
    :param keys: tuple of tuple of keys for the hash functions as bytes
    :param xor_folds: number of XOR folds to perform
    :param l: length of the Bloom filter in number of bits
    :param k: number of hash functions to use per element
    :param ngram_encoding:

    :return: 3-tuple:
            - bloom filter for record as a bitarray
            - first element of record (usually an index)
            - number of bits set in the bloomfilter
    """
    bloomfilter = bitarray(l)
    bloomfilter.setall(False)

    for (entry, tokenizer, f_keys) in zip(record, tokenizers, keys):
        ngrams = [ngram for ngram in tokenizer(str(entry))]
        if tokenizer.weight < 0:
            raise ValueError('weight must not be smaller than zero, but was: {}'.format(tokenizer.weight))
        adjusted_k = int(round(tokenizer.weight * k))
        bloomfilter |= ngram_encoding(ngrams, f_keys, adjusted_k, l)

    bloomfilter = fold_xor(bloomfilter, xor_folds)

    return bloomfilter, record[0], bloomfilter.count()


def stream_bloom_filters(dataset,       # type: Iterable[Tuple[Any, ...]]
                         schema_types,  # type: Iterable[IdentifierType]
                         keys,          # type: Tuple[Tuple[bytes, ...]]
                         xor_folds=0    # type: int
                         ):
    # type: (...) -> Iterable[Tuple[bitarray, Any, int]]
    """
    Yield bloom filters

    :param dataset: An iterable of indexable records.
    :param schema_types: An iterable of identifier type names.
    :param keys: A tuple of two lists of secret keys used in the HMAC.
    :param xor_folds: number of XOR folds to perform
    :return: Yields bloom filters as 3-tuples
    """
    for s in dataset:
        yield crypto_bloom_filter(s, schema_types, keys, xor_folds=xor_folds)


def calculate_bloom_filters(dataset,     # type: Iterable[Tuple[Any]]
                            schema,      # type: Iterable[IdentifierType]
                            keys,        # type: Tuple[Tuple[bytes, ...]]
                            xor_folds=0  # type: int
                            ):
    # type: (...) -> List[Tuple[bitarray, Any, int]]
    """
    :param dataset: A list of indexable records.
    :param schema: An iterable of identifier types.
    :param keys: A tuple of two lists of secret keys used in the HMAC.
    :param xor_folds: number of XOR folds to perform
    :return: List of bloom filters as 3-tuples, each containing
             bloom filter (bitarray), record first element - usually index, bitcount (int)
    """
    return list(stream_bloom_filters(dataset, schema, keys, xor_folds=xor_folds))


def serialize_bitarray(ba):
    # type: (bitarray) -> str
    """Serialize a bitarray (bloomfilter)

    """
    return base64.b64encode(ba.tobytes()).decode('utf8')



"""
Functions to tokenize words (PII)
"""
from typing import Optional, List


def unigramlist(instr, toremove=None, positional=False):
    # type: (str, Optional[List[str]], bool) -> List[str]
    """
    Make 1-grams (unigrams) from a word, possibly excluding particular substrings

    :param instr: input string
    :param toremove: Iterable of strings to remove
    :return: list of strings with unigrams
    """
    if toremove is not None:
        for substr in toremove:
            instr = instr.replace(substr, "")

    if positional:
        return positional_unigrams(instr)
    else:
        return list(instr)


def bigramlist(word, toremove=None):
    # type: (str, Optional[List[str]]) -> List[str]
    """
    Make bigrams from word with pre- and ap-pended spaces

    s -> [' ' + s0, s0 + s1, s1 + s2, .. sN + ' ']

    :param word: string to make bigrams from
    :param toremove: List of strings to remove before construction
    :return: list of bigrams as strings
    """
    if toremove is not None:
        for substr in toremove:
            word = word.replace(substr, "")
    word = " " + word + " "
    return [word[i:i+2] for i in range(len(word)-1)]


def positional_unigrams(instr):
    # type: (str) -> List[str]
    """
    Make positional unigrams from a word.

    E.g. 1987 -> ["1 1", "2 9", "3 8", "4 7"]

    :param instr: input string
    :return: list of strings with unigrams
    """
    return ["{index} {value}".format(index=i, value=c) for i, c in enumerate(instr, start=1)]



#!/usr/bin/env python3.6
from __future__ import print_function

import json
import time

import click
import requests

import clkhash
from clkhash import clk
from clkhash import benchmark as bench
from clkhash import randomnames
from clkhash.schema import get_schema_types, load_schema


DEFAULT_SERVICE_URL = 'https://es.data61.xyz'


def log(m, color='red'):
    click.echo(click.style(m, fg=color), err=True)


@click.group("clkutil")
@click.version_option(clkhash.__version__)
@click.option('--verbose', '-v', is_flag=True,
              help='Enables verbose mode.')
def cli(verbose=False):
    """
    This command line application allows a user to hash their
    data into cryptographic longterm keys for use in
    private comparison.

    This tool can also interact with a entity matching service;
    creating new mappings, uploading locally hashed data,
    watching progress, and retrieving results.

    Example:

        clkutil hash private_data.csv secretkey1 secretkey2 output-clks.json


    All rights reserved Confidential Computing 2016.
    """



@cli.command('hash', short_help="generate hashes from local PII data")
@click.argument('input', type=click.File('r'))
@click.argument('keys', nargs=2, type=click.Tuple([str, str]))
@click.option('--schema', '-s', type=click.File('r'), default=None)
@click.argument('output', type=click.File('w'))
@click.option('-q', '--quiet', default=False, is_flag=True, help="Quiet any progress messaging")
@click.option('--no-header', default=False, is_flag=True, help="Don't skip the first row")
@click.option('--xor-folds', default=0, type=click.IntRange(0, None))
def hash(input, output, schema, keys, quiet, no_header, xor_folds):
    """Process data to create CLKs

    Given a file containing CSV data as INPUT, and optionally a json
    document defining the expected schema, verify the schema, then
    hash the data to create CLKs writing to OUTPUT. Note the CSV
    file should contain a header row - however this row is not used
    by this tool.

    It is important that the keys are only known by the two data providers. Two words should be provided. For example:

    $clkutil hash input.txt horse staple output.txt

    Use "-" to output to stdout.
    """

    schema_types = get_schema_types(load_schema(schema))

    clk_data = clk.generate_clk_from_csv(input, keys, schema_types, no_header, not quiet, xor_folds)
    json.dump({'clks': clk_data}, output)
    if hasattr(output, 'name'):
        log("CLK data written to {}".format(output.name))


@cli.command('status', short_help='Get status of entity service')
@click.option('--server', type=str, default=DEFAULT_SERVICE_URL, help="Server address including protocol")
@click.option('-o','--output', type=click.File('w'), default='-')
@click.option('-v', '--verbose', default=False, is_flag=True, help="Script is more talkative")
def status(server, output, verbose):
    """Connect to an entity matching server and check the service status.

    Use "-" to output status to stdout.
    """
    log("Connecting to Entity Matching Server: {}".format(server))

    response = requests.get(server + "/api/v1/status")
    server_status = response.json()
    log("Response: {}".format(response.status_code))
    log("Status: {}".format(server_status['status']))
    print(json.dumps(server_status), file=output)


MAPPING_CREATED_MSG = """
The generated tokens can be used to upload hashed data and
fetch the resulting linkage table from the service.

To upload using the cli tool for entity A:

    clkutil hash a_people.csv A_HASHED_FILE.json
    clkutil upload --mapping="{resource_id}" --apikey="{update_tokens[0]}"  A_HASHED_FILE.json

To upload using the cli tool for entity B:

    clkutil hash b_people.csv B_HASHED_FILE.json
    clkutil upload --mapping="{resource_id}" --apikey="{update_tokens[1]}" B_HASHED_FILE.json

After both users have uploaded their data one can watch for and retrieve the results:

    clkutil results -w --mapping="{resource_id}" --apikey="{result_token}" --output results.txt

"""

@cli.command('create', short_help="create a mapping on the entity service")
@click.option('--type', default='permutation_unencrypted_mask',
              help='Alternative protocol/view type of the mapping. Default is unencrypted permutation and mask.')
@click.option('--schema', type=click.File('r'), help="Schema to publicly share with participating parties.")
@click.option('--server', type=str, default=DEFAULT_SERVICE_URL, help="Server address including protocol")
@click.option('-o','--output', type=click.File('w'), default='-')
@click.option('-t','--threshold', type=float, default=0.95)
@click.option('-v', '--verbose', default=False, is_flag=True, help="Script is more talkative")
def create(type, schema, server, output, threshold, verbose):
    """Create a new mapping on an entity matching server.

    See entity matching service documentation for details on mapping type, threshold
    and schema.

    Returns authentication details for the created mapping.
    """

    log("Entity Matching Server: {}".format(server))

    log("Checking server status")
    status = requests.get(server + "/api/v1/status").json()['status']
    log("Server Status: {}".format(status))

    if schema is not None:
        schema_object = load_schema(schema)
        schema_json = json.dumps(schema_object)
    else:
        schema_json = 'NOT PROVIDED'

    log("Schema: {}".format(schema_json))
    log("Type: {}".format(type))

    log("Creating new mapping")
    response = requests.post(
        "{}/api/v1/mappings".format(server),
        json={
            'schema': schema_json,
            'result_type': type,
            'threshold': threshold
        }
    )

    if response.status_code != 200:
        log("Unexpected response")
        log(response.text)
    else:
        log("Mapping created")
        if verbose:
            log(MAPPING_CREATED_MSG.format(**response.json()))
        print(response.text, file=output)


@cli.command('upload', short_help='upload hashes to entity service')
@click.argument('input', type=click.File('r'))
@click.option('--mapping', help='Server identifier of the mapping')
@click.option('--apikey', help='Authentication API key for the server.')
@click.option('--server', type=str, default=DEFAULT_SERVICE_URL, help="Server address including protocol")
@click.option('-o','--output', type=click.File('w'), default='-')
@click.option('-v', '--verbose', default=False, is_flag=True, help="Script is more talkative")
def upload(input, mapping, apikey, server, output, verbose):
    """Upload CLK data to entity matching server.

    Given a json file containing hashed clk data as INPUT, upload to
    the entity resolution service.

    Use "-" to read from stdin.
    """

    log("Uploading CLK data from {}".format(input.name))
    log("To Entity Matching Server: {}".format(server))
    log("Mapping ID: {}".format(mapping))

    log("Checking server status")
    status = requests.get(server + "/api/v1/status").json()['status']
    log("Status: {}".format(status))

    log("Uploading CLK data to the server")

    response = requests.put(
        '{}/api/v1/mappings/{}'.format(server, mapping),
        data=input,
        headers={
            "Authorization": apikey,
            'content-type': 'application/json'
        }
    )

    if verbose:
        log(response.text)
        log("When the other party has uploaded their CLKS, you should be able to watch for results")


    print(response.text, file=output)



@cli.command('results', short_help="fetch results from entity service")
@click.option('--mapping',
              help='Server identifier of the mapping')
@click.option('--apikey', help='Authentication API key for the server.')
@click.option('-w', '--watch', help='Follow/wait until results are available', is_flag=True)
@click.option('--server', type=str, default=DEFAULT_SERVICE_URL, help="Server address including protocol")
@click.option('-o','--output', type=click.File('w'), default='-')
def results(mapping, apikey, watch, server, output):
    """
    Check to see if results are available for a particular mapping
    and if so download.

    Authentication is carried out using the --apikey option which
    must be provided. Depending on the server operating mode this
    may return a mask, a linkage table, or a permutation. Consult
    the entity service documentation for details.
    """

    log("Checking server status")
    status = requests.get(server + "/api/v1/status").json()['status']
    log("Status: {}".format(status))

    def get_result():
        return requests.get(
            '{}/api/v1/mappings/{}'.format(server, mapping),
            headers={"Authorization": apikey}
        )

    response = get_result()
    log("Response code: {}".format(response.status_code))

    if watch:
        while response.status_code != 200:
            time.sleep(1)
            response = get_result()

    if response.status_code == 200:
        log("Received result")
        print(response.text, file=output)
    else:
        log("No result yet")
        log(response.text)


@cli.command('benchmark', short_help='carry out a local benchmark')
def benchmark():
    bench.compute_hash_speed(10000)


@cli.command('generate', short_help='generate random pii data for testing')
@click.argument('size', type=int, default=100)
@click.argument('output', type=click.File('w'))
@click.option('--schema', '-s', type=click.File('r'), default=None)
def generate(size, output, schema):
    """Generate fake PII data for testing"""
    pii_data = randomnames.NameList(size)

    if schema is not None:
        raise NotImplementedError

    randomnames.save_csv(pii_data.names, pii_data.schema, output)


if __name__ == "__main__":
    cli()



from setuptools import setup, find_packages
import codecs

requirements = [
        "bitarray==0.8.1",
        "click==6.7",
        "requests==2.18.4",
        "futures>=3.1; python_version == '2.7'",
        "cryptography==2.1.3",
        "tqdm==4.19.4",
        "typing>=3.6.2",
        "future==0.16.0",
        "pyblake2==1.1.0; python_version < '3.6'",
        "enum34==1.1.6; python_version < '3.4'"
    ]

with codecs.open('README.md', 'r', 'utf-8') as f:
    readme = f.read()

setup(
    name="clkhash",
    version='0.10.1',
    description='Hash utility to create Cryptographic Linkage Keys',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/n1analytics/clkhash',
    license='Apache',
    install_requires=requirements,
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-cov'],
    packages=find_packages(exclude=['tests']),
    package_data={'clkhash': ['data/*.csv']},
    entry_points={
        'console_scripts': [
            'clkutil = clkhash.cli:cli'
        ],
    }
)



"""
Convert PII to tokens
"""
from typing import Dict, List, Any

from clkhash.tokenizer import unigramlist, bigramlist
from copy import copy


class IdentifierType:
    """
    Base class used for all identifier types.

    Required to provide a mapping of schema to hash type
    uni-gram or bi-gram.
    """

    def __init__(self, unigram=False, weight=1, **kwargs):
        # type: (bool, float, Any) -> None
        """
        :param bool unigram: Use uni-gram instead of using bi-grams
        :param float weight: adjusts the "importance" of this identifier in the Bloom filter. Can be set to zero to skip
        :param kwargs: Extra keyword arguments passed to the tokenizer

        .. Note::
           For each n-gram of an identifier, we compute *k* different indices in the Bloom filter which will be set to
           true. There is a global :math:`k_{default}` value, and the *k* value for each identifier is computed as

           .. math::
              k = weight * k_{default},

           rounded to the nearest integer.

           Reasons why you might want to set weights:

             - Long identifiers like street name will produce a lot more n-grams than small identifiers like zip code.
               Thus street name will flip more bits in the Bloom filter and will have a bigger influence in the overall
               matching score.

             - The matching might produce better results if identifiers that are stable and / or have low error rates
               are given higher prominence in the Bloom filter.

        """
        self.weight = weight
        self.tokenizer = unigramlist if unigram else bigramlist
        self.kwargs = kwargs

    def __call__(self, entry):
        # type: (str) -> List[str]
        return self.tokenizer(entry, **self.kwargs)  # type: ignore


basic_types = {
    'INDEX': IdentifierType(weight=0),

    'GENDER M or F': IdentifierType(unigram=True),
    'GENDER freetext': IdentifierType(),

    'DOB YYYY/MM/DD': IdentifierType(toremove='/'),
    'DOB YYYY': IdentifierType(unigram=True, toremove='/'),

    'NAME freetext': IdentifierType(),

    'ANY freetext': IdentifierType(),

    'PHONE freetext': IdentifierType(unigram=True, toremove='()-')
}

# Weighted Types
# Zip Code, Birth Year, Birth Month, Birth Day, Sex, and House number
# can be regarding as identifiers with low error rates and First name,
# last name, street name, place name are in most applications found to
# be more error prone.


weighted_types = {
    'INDEX': IdentifierType(weight=0),

    # gender weight = 1 due to lower identifier entropy
    'GENDER M or F': IdentifierType(unigram=True),

    'DOB DD': IdentifierType(unigram=True, weight=2),
    'DOB MM': IdentifierType(unigram=True, weight=2),
    'DOB YYYY': IdentifierType(unigram=True, toremove='/', weight=2, positional=True),

    'ADDRESS House Number': IdentifierType(weight=2),
    'ADDRESS POSTCODE': IdentifierType(unigram=True, weight=2, positional=True),
    'ADDRESS Place Name': IdentifierType(weight=1),

    'NAME First Name': IdentifierType(),
    'NAME Surname': IdentifierType(),

    'PHONE freetext': IdentifierType(unigram=True, toremove='()-')
}


def identifier_type_from_description(schema_object):
    # type: (Dict[str, Any]) -> IdentifierType
    """
    Convert a dictionary describing a feature into an IdentifierType

    :param schema_object:
    :return: An IdentifierType
    """
    assert "identifier" in schema_object
    id = schema_object['identifier']

    if id in basic_types:
        id_type = basic_types[id]
    elif id in weighted_types:
        id_type = weighted_types[id]
    else:
        id_type = IdentifierType(
            weight=schema_object.get('weight', 1)
        )

    # check if there was a custom weight
    if 'weight' in schema_object:
        id_type = copy(id_type)  # we don't want to modify the original
        id_type.weight = schema_object['weight']

    return id_type
